/////////////////////////////////////////////////////////////////////////////
// JR 東日本トレインシミュレーター コントローラーのテスト
//
// Copyright (C) 2023, Kentaro HARA.
/////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "hardware/gpio.h"
#include "pico/binary_info.h"
#include "bsp/board.h"
#include "tusb.h"

#include "usb_descriptors.h"

//--------------------------------------------------------------------+
// MACRO CONSTANT TYPEDEF PROTYPES
//--------------------------------------------------------------------+

/* Blink pattern
 * - 250 ms  : device not mounted
 * - 1000 ms : device mounted
 * - 2500 ms : device is suspended
 */
enum
{
    BLINK_NOT_MOUNTED = 250,
    BLINK_MOUNTED = 1000,
    BLINK_SUSPENDED = 2500,
};

static uint32_t blink_interval_ms = BLINK_NOT_MOUNTED;
static uint s_encoderPrev = 0;      // ロータリー エンコーダの前回値
static uint s_input = 0;            // ボタンの入力

uint ReadEncoder(void);
uint GetEncoderStatus(void);
uint GetButtonStatus(void);
uint GetInput(void);
void led_blinking_task(void);
void hid_task(void);

// GPIO ピン/ボタン ID の定義
enum
{
    LED_PIN = 25,       // LED
    ENCODER_A = 2,      // ロータリー エンコーダ A 相
    ENCODER_B = 3,      // ロータリー エンコーダ B 相
    BUTTON_ACCEL = 101, // 加速強め/ブレーキ弱め
    BUTTON_BRAKE = 102, // 加速弱め/ブレーキ強め
    BUTTON_NEUTRAL = 4, // ニュートラル
    BUTTON_ENTER = 6,   // [Enter] キー
    BUTTON_ESC = 7,     // [Esc] キー
    BUTTON_UP = 8,      // [↑] キー
    BUTTON_DOWN = 9,    // [↓] キー
    BUTTON_RIGHT = 10,  // [→] キー
    BUTTON_LEFT = 11    // [←] キー
};

// ロータリー エンコーダ回転方向定義
enum EndocerRotDir
{
    ENCODER_ROT_NONE,   // 回転なし
    ENCODER_ROT_CW,     // 時計回り
    ENCODER_ROT_CCW     // 反時計回り
};

int main()
{
    // 各種初期化を行います。
    board_init();
    tusb_init();
    stdio_init_all();

    // GPIO を初期化します。
    gpio_init(LED_PIN);
    gpio_init(ENCODER_A);
    gpio_init(ENCODER_B);
    gpio_init(BUTTON_NEUTRAL);
    gpio_init(BUTTON_ENTER);
    gpio_init(BUTTON_ESC);
    gpio_init(BUTTON_UP);
    gpio_init(BUTTON_DOWN);
    gpio_init(BUTTON_RIGHT);
    gpio_init(BUTTON_LEFT);

    // GPIO の入出力方向を設定します。
    gpio_set_dir(LED_PIN, GPIO_OUT);
    gpio_set_dir(ENCODER_A, GPIO_IN);
    gpio_set_dir(ENCODER_B, GPIO_IN);
    gpio_set_dir(BUTTON_NEUTRAL, GPIO_IN);
    gpio_set_dir(BUTTON_ENTER, GPIO_IN);
    gpio_set_dir(BUTTON_ESC, GPIO_IN);
    gpio_set_dir(BUTTON_UP, GPIO_IN);
    gpio_set_dir(BUTTON_DOWN, GPIO_IN);
    gpio_set_dir(BUTTON_RIGHT, GPIO_IN);
    gpio_set_dir(BUTTON_LEFT, GPIO_IN);

    // 入力端子をプルアップします。
    gpio_pull_up(ENCODER_A);
    gpio_pull_up(ENCODER_B);
    gpio_pull_up(BUTTON_NEUTRAL);
    gpio_pull_up(BUTTON_ENTER);
    gpio_pull_up(BUTTON_ESC);
    gpio_pull_up(BUTTON_UP);
    gpio_pull_up(BUTTON_DOWN);
    gpio_pull_up(BUTTON_RIGHT);
    gpio_pull_up(BUTTON_LEFT);

    while (1)
    {
        tud_task(); // tinyusb device task
        led_blinking_task();
        hid_task();
    }

    return 0;
}

/*! \brief ロータリー エンコーダを読み取ります。
 * \return bit0 が A 相のレベル、bit1 が B 相のレベル。
 */
uint ReadEncoder(void)
{
    uint ret = 0;

    ret |= gpio_get(ENCODER_A) ? 1 : 0;
    ret |= gpio_get(ENCODER_B) ? 2 : 0;

    return ret;
}

/*! \brief 押されたボタンを返します (ボタンはロー アクティブ)。
 * \return 押されたボタンの ID。ボタンが何も押されていなければ、0。
 */
uint GetButtonStatus(void)
{
    const uint buttons[] =
        {
            BUTTON_NEUTRAL,
            BUTTON_ENTER,
            BUTTON_ESC,
            BUTTON_UP,
            BUTTON_DOWN,
            BUTTON_RIGHT,
            BUTTON_LEFT
        };

    // 入力ポートをスキャンします。
    for (int i = 0; i < count_of(buttons); i++)
    {
        if (!gpio_get(buttons[i]))
        {
            return buttons[i];
        }
    }

    return 0;
}

/*! \brief ロータリー エンコーダの回転方向を返します。
 * \return EncoderRotDir 列挙体の値のいずれか。
 */
uint GetEncoderStatus(void)
{
    uint enc = ReadEncoder();
    uint status = ENCODER_ROT_NONE;

    // 回転した場合 (A 相の立ち下がり)、
    if ((s_encoderPrev & 1) && (!(enc & 1)))
    {
        // B 相のレベルが 1 なら CW、0 なら CCW。
        if (enc & 2)
        {
            status = ENCODER_ROT_CW;
        }
        else
        {
            status = ENCODER_ROT_CCW;
        }
    }

    s_encoderPrev = enc;
    return status;
}

/*! \brief 押されたボタン/ロータリー エンコーダの ID を返します。
 * \return 押されたボタン/ロータリー エンコーダの ID。入力がなければ 0。
 */
uint GetInput(void)
{
    uint status = GetButtonStatus();

    if (status)
    {
        return status;
    }
    else
    {
        switch (GetEncoderStatus())
        {
        case ENCODER_ROT_CW:
            return BUTTON_ACCEL;    // 加速
            break;
        
        case ENCODER_ROT_CCW:
            return BUTTON_BRAKE;    // 減速
            break;

        default:
            return 0;
            break;
        }
    }
}

//--------------------------------------------------------------------+
// Device callbacks
//--------------------------------------------------------------------+

// Invoked when device is mounted
void tud_mount_cb(void)
{
    blink_interval_ms = BLINK_MOUNTED;
}

// Invoked when device is unmounted
void tud_umount_cb(void)
{
    blink_interval_ms = BLINK_NOT_MOUNTED;
}

// Invoked when usb bus is suspended
// remote_wakeup_en : if host allow us  to perform remote wakeup
// Within 7ms, device must draw an average of current less than 2.5 mA from bus
void tud_suspend_cb(bool remote_wakeup_en)
{
    (void)remote_wakeup_en;
    blink_interval_ms = BLINK_SUSPENDED;
}

// Invoked when usb bus is resumed
void tud_resume_cb(void)
{
    blink_interval_ms = BLINK_MOUNTED;
}

//--------------------------------------------------------------------+
// USB HID
//--------------------------------------------------------------------+

static void send_hid_report(uint8_t report_id, uint32_t btn)
{
    // skip if hid is not ready yet
    if (!tud_hid_ready())
        return;

    switch (report_id)
    {
    case REPORT_ID_KEYBOARD:
    {
        // use to avoid send multiple consecutive zero report for keyboard
        // キーボードに対して複数の連続したゼロ レポートが送信されるのを避けるために使用します
        static bool has_keyboard_key = false;
        uint8_t keycode[6] = {0};

        switch (btn)
        {
        case BUTTON_ACCEL:
        case BUTTON_NEUTRAL:
        case BUTTON_BRAKE:
            keycode[0] = HID_KEY_NONE;  // これらはマウス レポートを送信するため、キーは送信しません。
            break;

        case BUTTON_ENTER:
            keycode[0] = HID_KEY_ENTER;
            break;

        case BUTTON_ESC:
            keycode[0] = HID_KEY_ESCAPE;
            break;

        case BUTTON_UP:
            keycode[0] = HID_KEY_ARROW_UP;
            break;

        case BUTTON_DOWN:
            keycode[0] = HID_KEY_ARROW_DOWN;
            break;

        case BUTTON_RIGHT:
            keycode[0] = HID_KEY_ARROW_RIGHT;
            break;

        case BUTTON_LEFT:
            keycode[0] = HID_KEY_ARROW_LEFT;
            break;

        default:
            break;
        }

        if (btn)
        {
            tud_hid_keyboard_report(REPORT_ID_KEYBOARD, 0, keycode);
            has_keyboard_key = true;
        }
        else
        {
            // send empty key report if previously has key pressed
            // 以前にキーが押された場合は空のキーのレポートを送信する
            if (has_keyboard_key)
                tud_hid_keyboard_report(REPORT_ID_KEYBOARD, 0, NULL);
            has_keyboard_key = false;
        }
    }
    break;

    case REPORT_ID_MOUSE:
    {
        static bool buttonPressed = false;

        switch (btn)
        {
        case BUTTON_ACCEL:
            // ホイール下回転
            tud_hid_mouse_report(REPORT_ID_MOUSE, 0x00, 0, 0, -1, 0);
            break;
        
        case BUTTON_BRAKE:
            // ホイール上回転
            tud_hid_mouse_report(REPORT_ID_MOUSE, 0x00, 0, 0, 1, 0);
            break;

        case BUTTON_NEUTRAL:
            // ホイール クリック
            tud_hid_mouse_report(REPORT_ID_MOUSE, MOUSE_BUTTON_MIDDLE, 0, 0, 0, 0);
            buttonPressed = true;
            break;

        default:
            if (buttonPressed)
            {
                tud_hid_mouse_report(REPORT_ID_MOUSE, 0x00, 0, 0, 0, 0);
            }
            buttonPressed = false;
            break;
        }
    }
    break;

    // case REPORT_ID_CONSUMER_CONTROL:
    // {
    //     // use to avoid send multiple consecutive zero report
    //     static bool has_consumer_key = false;

    //     if (btn)
    //     {
    //         // volume down
    //         uint16_t volume_down = HID_USAGE_CONSUMER_VOLUME_DECREMENT;
    //         tud_hid_report(REPORT_ID_CONSUMER_CONTROL, &volume_down, 2);
    //         has_consumer_key = true;
    //     }
    //     else
    //     {
    //         // send empty key report (release key) if previously has key pressed
    //         uint16_t empty_key = 0;
    //         if (has_consumer_key)
    //             tud_hid_report(REPORT_ID_CONSUMER_CONTROL, &empty_key, 2);
    //         has_consumer_key = false;
    //     }
    // }
    // break;

    // case REPORT_ID_GAMEPAD:
    // {
    //     // use to avoid send multiple consecutive zero report for keyboard
    //     static bool has_gamepad_key = false;

    //     hid_gamepad_report_t report =
    //         {
    //             .x = 0, .y = 0, .z = 0, .rz = 0, .rx = 0, .ry = 0, .hat = 0, .buttons = 0};

    //     if (btn)
    //     {
    //         report.hat = GAMEPAD_HAT_UP;
    //         report.buttons = GAMEPAD_BUTTON_A;
    //         tud_hid_report(REPORT_ID_GAMEPAD, &report, sizeof(report));

    //         has_gamepad_key = true;
    //     }
    //     else
    //     {
    //         report.hat = GAMEPAD_HAT_CENTERED;
    //         report.buttons = 0;
    //         if (has_gamepad_key)
    //             tud_hid_report(REPORT_ID_GAMEPAD, &report, sizeof(report));
    //         has_gamepad_key = false;
    //     }
    // }
    // break;

    default:
        break;
    }
}

// Every 5ms, we will sent 1 report for each HID profile (keyboard, mouse etc ..)
// tud_hid_report_complete_cb() is used to send the next report after previous one is complete
void hid_task(void)
{
    // Poll every 5ms
    const uint32_t interval_ms = 5;
    static uint32_t start_ms = 0;

    if (board_millis() - start_ms < interval_ms)
        return; // not enough time
    start_ms += interval_ms;

    //uint32_t const btn = board_button_read();
    s_input = GetInput();

    // Remote wakeup
    if (tud_suspended() && s_input)
    {
        // Wake up host if we are in suspend mode
        // and REMOTE_WAKEUP feature is enabled by host
        tud_remote_wakeup();
    }
    else
    {
        // Send the 1st of report chain, the rest will be sent by tud_hid_report_complete_cb()
        send_hid_report(REPORT_ID_KEYBOARD, s_input);
    }
}

// Invoked when sent REPORT successfully to host
// Application can use this to send the next report
// Note: For composite reports, report[0] is report ID
void tud_hid_report_complete_cb(uint8_t instance, uint8_t const *report, uint16_t len)
{
    (void)instance;
    (void)len;
    (void)report;

    uint8_t next_report_id = report[0] + 1;

    if (next_report_id <= REPORT_ID_MOUSE)
    {
        send_hid_report(next_report_id, s_input);
    }
}

// Invoked when received GET_REPORT control request
// Application must fill buffer report's content and return its length.
// Return zero will cause the stack to STALL request
uint16_t tud_hid_get_report_cb(uint8_t instance, uint8_t report_id, hid_report_type_t report_type, uint8_t *buffer, uint16_t reqlen)
{
    // TODO not Implemented
    (void)instance;
    (void)report_id;
    (void)report_type;
    (void)buffer;
    (void)reqlen;

    return 0;
}

// Invoked when received SET_REPORT control request or
// received data on OUT endpoint ( Report ID = 0, Type = 0 )
void tud_hid_set_report_cb(uint8_t instance, uint8_t report_id, hid_report_type_t report_type, uint8_t const *buffer, uint16_t bufsize)
{
    (void)instance;

    if (report_type == HID_REPORT_TYPE_OUTPUT)
    {
        // Set keyboard LED e.g Capslock, Numlock etc...
        if (report_id == REPORT_ID_KEYBOARD)
        {
            // bufsize should be (at least) 1
            if (bufsize < 1)
                return;

            uint8_t const kbd_leds = buffer[0];

            if (kbd_leds & KEYBOARD_LED_CAPSLOCK)
            {
                // Capslock On: disable blink, turn led on
                blink_interval_ms = 0;
                board_led_write(true);
            }
            else
            {
                // Caplocks Off: back to normal blink
                board_led_write(false);
                blink_interval_ms = BLINK_MOUNTED;
            }
        }
    }
}

//--------------------------------------------------------------------+
// BLINKING TASK
//--------------------------------------------------------------------+
void led_blinking_task(void)
{
    static uint32_t start_ms = 0;
    static bool led_state = false;

    // blink is disabled
    if (!blink_interval_ms)
        return;

    // Blink every interval ms
    if (board_millis() - start_ms < blink_interval_ms)
        return; // not enough time
    start_ms += blink_interval_ms;

    board_led_write(led_state);
    led_state = 1 - led_state; // toggle
}
